#include <ros/ros.h>
#include <teleop_msgs/body_angle.h>
#include <geometry_msgs/Point.h>
#include <std_msgs/String.h>
#include <std_msgs/Bool.h>
#include <teleop_msgs/joint_angle.h>
#include <teleop_msgs/walk_info.h>

int estado = 0; //Estaddo del controlador

std_msgs::String orden; //Ultima palabra dicha por el usuario

geometry_msgs::Point posCabeza; //posicion inicial de la cabeza

//Declaracion de los publicadores
ros::Publisher naoDicePub;

ros::Publisher aperturaHombroDPub;
ros::Publisher rotacionHombroDPub;
ros::Publisher aperturaCodoDPub;
ros::Publisher rotacionCodoDPub;
ros::Publisher rotacionManoDPub;
ros::Publisher aperturaManoDPub;

ros::Publisher rotacionCabezaPub;
ros::Publisher inclinacionCabezaPub;

ros::Publisher aperturaHombroIPub;
ros::Publisher rotacionHombroIPub;
ros::Publisher aperturaCodoIPub;
ros::Publisher rotacionCodoIPub;
ros::Publisher rotacionManoIPub;
ros::Publisher aperturaManoIPub;

ros::Publisher walkInfoPub;
//

//Declaracion de los mensajes
teleop_msgs::joint_angle aperturaHombroDMsg;
teleop_msgs::joint_angle rotacionHombroDMsg;
teleop_msgs::joint_angle aperturaCodoDMsg;
teleop_msgs::joint_angle rotacionCodoDMsg;
teleop_msgs::joint_angle rotacionManoDMsg;

teleop_msgs::joint_angle rotacionCabezaMsg;
teleop_msgs::joint_angle inclinacionCabezaMsg;

teleop_msgs::joint_angle aperturaHombroIMsg;
teleop_msgs::joint_angle rotacionHombroIMsg;
teleop_msgs::joint_angle aperturaCodoIMsg;
teleop_msgs::joint_angle rotacionCodoIMsg;
teleop_msgs::joint_angle rotacionManoIMsg;

teleop_msgs::walk_info walkInfoMsg;
//

#define PI 3.14159265358979323846

/*
 * Funcion que calcula como debe andar el NAO a partir de la posicion del cuerpo
 */
void calcularAndar(const teleop_msgs::body_angle bodyAngle)
{
  double difX = posCabeza.x - bodyAngle.Head.x; //Calcula la distancia en x que se ha movido la cabeza
  double difZ = posCabeza.z - bodyAngle.Head.z; //Calcula la distancia en z que se ha movido la cabeza
  double difZHombros = bodyAngle.RShoulderPosition.z - bodyAngle.LShoulderPosition.z; //Calcula la diferencia en z de los dos hombros

  //Inicializa el mensaje
  walkInfoMsg.angle = 0;
  walkInfoMsg.x = 0;
  walkInfoMsg.y = 0;

  //Calculo del giro
  //Si la distancia entre la coordenadas z de los hombros en mayor de 15
  if (std::abs(difZHombros) > 0.15)
  {
    if (difZHombros < 0)
      walkInfoMsg.angle = 0.2; //Se gira hacia un lado
    else
      walkInfoMsg.angle = -0.2; //Se gira hacia el otro
  }
  //Calculo de la direccion
  if (std::abs(difZ) > std::abs(difX)) //Si la persona se ha movido mas hacia delante o atras que hacia los lados
  {
    if (std::abs(difZ) > 0.2)
    {
      if (difZ < 0)
        walkInfoMsg.x = -0.4; //Se anda hacia atras
      else
        walkInfoMsg.x = 0.4; //Se anda hacia delante

    }
  }
  else
  {
    if (std::abs(difX) > 0.20)
    {
      if (difX < 0)
        walkInfoMsg.y = -0.4; //Se anda hacia la izquierda
      else
        walkInfoMsg.y = 0.4; //Se anda hacia la derecha
    }
  }
}

/*
 * Funcion que a partir de la informacion del cuerpo recibida, crea los mensajes que se le manadaran al nao para que se mueva
 */
void calcularAngulosNao(const teleop_msgs::body_angle bodyAngle)
{
  //Algunos de los angulos recibidos necesitan una transformacion para ser comunicados al NAO.
  aperturaHombroDMsg.angle = bodyAngle.RShoulderOpeningAngle * -1;
  rotacionHombroDMsg.angle = (bodyAngle.RShoulderRotationAngle - PI / 2) * -1;
  aperturaCodoDMsg.angle = bodyAngle.RElbowOpeningAngle;
  rotacionCodoDMsg.angle = bodyAngle.RElbowRotationAngle;
  rotacionManoDMsg.angle = bodyAngle.RHandRotationAngle;

  //Se calcula el angulo de la cabeza para que la camara siga a la mano cuando va a agarrar algo
  if (bodyAngle.RShoulderRotationAngle > 50 * PI / 180 && bodyAngle.RShoulderRotationAngle < 105 * PI / 180
      && bodyAngle.RElbowOpeningAngle < 45 * PI / 180)
  {
    double a = sqrt(pow(0.1, 2) + pow(0.2, 2) - 2 * 0.11 * 0.2 * cos(bodyAngle.RShoulderOpeningAngle + PI / 2));
    double anguloRotacionAux = asin(0.2 * sin(bodyAngle.RShoulderOpeningAngle + PI / 2) / a);
    rotacionCabezaMsg.angle = -(PI / 2 - anguloRotacionAux);
    double b = sqrt(pow(0.07, 2) + pow(0.2, 2) - 2 * 0.07 * 0.2 * cos(PI - bodyAngle.RShoulderRotationAngle));
    double anguloInclinacionAux = asin(0.2 * sin(PI - bodyAngle.RShoulderRotationAngle) / b);
    inclinacionCabezaMsg.angle = (PI / 2 - anguloInclinacionAux) - 65 * PI / 180;
  }
  else if (bodyAngle.LShoulderRotationAngle > 50 * PI / 180 && bodyAngle.LShoulderRotationAngle < 105 * PI / 180
      && bodyAngle.LElbowOpeningAngle < 45 * PI / 180)
  {
    double a = sqrt(pow(0.1, 2) + pow(0.2, 2) - 2 * 0.1 * 0.2 * cos(bodyAngle.LShoulderOpeningAngle + PI / 2));
    double anguloRotacionAux = asin(0.2 * sin(bodyAngle.LShoulderOpeningAngle + PI / 2) / a);
    rotacionCabezaMsg.angle = (PI / 2 - anguloRotacionAux);
    double b = sqrt(pow(0.07, 2) + pow(0.2, 2) - 2 * 0.07 * 0.2 * cos(PI - bodyAngle.LShoulderRotationAngle));
    double anguloInclinacionAux = asin(0.2 * sin(PI - bodyAngle.LShoulderRotationAngle) / b);
    inclinacionCabezaMsg.angle = (PI / 2 - anguloInclinacionAux) - 65 * PI / 180;
  }
  else
  {
    rotacionCabezaMsg.angle = 0;
    inclinacionCabezaMsg.angle = -30 * PI / 180;
  }

  aperturaHombroIMsg.angle = bodyAngle.LShoulderOpeningAngle;
  rotacionHombroIMsg.angle = (bodyAngle.LShoulderRotationAngle - PI / 2) * -1;
  aperturaCodoIMsg.angle = bodyAngle.LElbowOpeningAngle * -1;
  rotacionCodoIMsg.angle = bodyAngle.LElbowRotationAngle * -1;
  rotacionManoIMsg.angle = bodyAngle.LHandRotationAngle;
}

void publicarPararNAO()
{
  //Inicializa el mensaje
  walkInfoMsg.angle = 0;
  walkInfoMsg.x = 0;
  walkInfoMsg.y = 0;
  walkInfoPub.publish(walkInfoMsg);
}

/*
 * Funcion encargada de publicar los mensajes
 */
teleop_msgs::walk_info walkInfoMsgAnterior;
void publicar()
{
  aperturaHombroDPub.publish(aperturaHombroDMsg);
  rotacionHombroDPub.publish(rotacionHombroDMsg);
  aperturaCodoDPub.publish(aperturaCodoDMsg);
  rotacionCodoDPub.publish(rotacionCodoDMsg);
  rotacionManoDPub.publish(rotacionManoDMsg);

  rotacionCabezaPub.publish(rotacionCabezaMsg);
  inclinacionCabezaPub.publish(inclinacionCabezaMsg);

  aperturaHombroIPub.publish(aperturaHombroIMsg);
  rotacionHombroIPub.publish(rotacionHombroIMsg);
  aperturaCodoIPub.publish(aperturaCodoIMsg);
  rotacionCodoIPub.publish(rotacionCodoIMsg);
  rotacionManoIPub.publish(rotacionManoIMsg);
  if (walkInfoMsg.x != walkInfoMsgAnterior.x || walkInfoMsg.y != walkInfoMsgAnterior.y
      || walkInfoMsg.angle != walkInfoMsgAnterior.angle)
  {
    walkInfoPub.publish(walkInfoMsg);
    walkInfoMsgAnterior.x = walkInfoMsg.x;
    walkInfoMsgAnterior.y = walkInfoMsg.y;
    walkInfoMsgAnterior.angle = walkInfoMsg.angle;

  }
}

bool aperturaManoD = 0; //mano derecha cerrada
bool aperturaManoI = 0; //mano izquierda cerrada

/*
 * Manejador que despierta cuando llega algun mensaje de audio, es decir, alguna palabra dicha por el usuario
 */
void audioHandle(const std_msgs::StringPtr &output)
{
  orden = *output;
  std_msgs::Bool aperturaManoIMsg, aperturaManoDMsg;
  if (estado == 1)
  {
    //Cuando el usuario dice "right", se cambia el estado de la mano derecha si estaba cerrada, pasa a abierta y viceversa.
    if (output->data.compare("right") == 0 || output->data.compare("derecha") == 0)
    {
      if (aperturaManoD == 0)
      {
        printf("Opening right\n");
        std_msgs::String naoDiceMsg;
        naoDiceMsg.data = "Opening right";
        naoDicePub.publish(naoDiceMsg);
      }
      else
      {
        printf("Closing the right\n");
        std_msgs::String naoDiceMsg;
        naoDiceMsg.data = "Closing the right";
        naoDicePub.publish(naoDiceMsg);
      }
      aperturaManoD = !aperturaManoD;
      aperturaManoDMsg.data = aperturaManoD;
      aperturaManoDPub.publish(aperturaManoDMsg);
      return;
    }
    //Cuando el usuario dice "left", se cambia el estado de la mano derecha si estaba cerrada, pasa a abierta y viceversa.
    if (output->data.compare("left") == 0 || output->data.compare("izquierda") == 0)
    {
      if (aperturaManoI == 0)
      {
        printf("Opening left\n");
        std_msgs::String naoDiceMsg;
        naoDiceMsg.data = "Opening left";
        naoDicePub.publish(naoDiceMsg);
      }
      else
      {
        printf("Closing the left\n");
        std_msgs::String naoDiceMsg;
        naoDiceMsg.data = "Closing the left";
        naoDicePub.publish(naoDiceMsg);
      }
      aperturaManoI = !aperturaManoI;
      aperturaManoIMsg.data = aperturaManoI;
      aperturaManoIPub.publish(aperturaManoIMsg);
      return;
    }
    if (orden.data.compare("finish") == 0 || orden.data.compare("parar") == 0)
    {
      estado = 0;
      publicarPararNAO();
      printf("NAO stopped.\n");
      printf("Say START to start.\n");
      std_msgs::String naoDiceMsg;
      naoDiceMsg.data = "Finish teleoperation";
      naoDicePub.publish(naoDiceMsg);
      return;
    }
  }
  if (estado == 0 && (orden.data.compare("start") == 0 || orden.data.compare("empezar") == 0))
  {
    printf("You can control the NAO.\n");
    std_msgs::String naoDiceMsg;
    naoDiceMsg.data = "Start teleoperation";
    naoDicePub.publish(naoDiceMsg);
    //Se cierran las manos
    aperturaManoI = 0;
    aperturaManoIMsg.data = 0;
    aperturaManoIPub.publish(aperturaManoIMsg);
    aperturaManoD = 0;
    aperturaManoDMsg.data = 0;
    aperturaManoDPub.publish(aperturaManoDMsg);
    estado = 1;
    return;
  }

  printf("Repeat please. I don't understand you.\n");
}

/*
 * Manejador que despierta cuando un mensaje con la informacion del cuerpo
 */
void bodyAngleHandle(const teleop_msgs::body_anglePtr &bodyAngle)
{

  switch (estado)
  {
    case 0: //Estado 0: Estado en el que no se esta controlando al Nao, a la espera de que el usuario diga Hello
      posCabeza = bodyAngle->Head; //Se memoriza la posicion de la cabeza para conocer la posicion inicial del cuerpo
      publicarPararNAO();
      break;
    case 1: //Estado 1: Se procede a manejar el Nao, si se dice "goodbye" se vuelve al estado 0.
      calcularAndar(*bodyAngle); //Se calcula como debe andar el robot
      calcularAngulosNao(*bodyAngle); //Se calcula el movimiento de los brazos
      publicar(); //Se publican los mensajes
      break;
  }
}

int main(int argc, char** argv)
{
  ros::init(argc, argv, "teleop_control");
  ros::NodeHandle n;
  sleep(15);
  printf("Calibrate your body and say START to start.\n");

  //Suscripcion a los topic
  ros::Subscriber bodyAngleSub, audioSub;
  bodyAngleSub = n.subscribe("/bodyAngle", 1, bodyAngleHandle); //Informacion del cuerpo
  audioSub = n.subscribe("/recognizer/output", 1, audioHandle); //Informacion del audio

  //SE anuncian los topic a publicar
  naoDicePub = n.advertise<std_msgs::String>("NaoSays", 1);

  aperturaHombroDPub = n.advertise<teleop_msgs::joint_angle>("RShoulderOpening", 1);
  rotacionHombroDPub = n.advertise<teleop_msgs::joint_angle>("RShoulderRotation", 1);
  aperturaCodoDPub = n.advertise<teleop_msgs::joint_angle>("RElbowOpening", 1);
  rotacionCodoDPub = n.advertise<teleop_msgs::joint_angle>("RElbowRotation", 1);
  rotacionManoDPub = n.advertise<teleop_msgs::joint_angle>("RHandRotation", 1);
  aperturaManoDPub = n.advertise<std_msgs::Bool>("RHandOpening", 1);

  inclinacionCabezaPub = n.advertise<teleop_msgs::joint_angle>("HeadPitch", 1);
  rotacionCabezaPub = n.advertise<teleop_msgs::joint_angle>("HeadRotation", 1);

  aperturaHombroIPub = n.advertise<teleop_msgs::joint_angle>("LShoulderOpening", 1);
  rotacionHombroIPub = n.advertise<teleop_msgs::joint_angle>("LShoulderRotation", 1);
  aperturaCodoIPub = n.advertise<teleop_msgs::joint_angle>("LElbowOpening", 1);
  rotacionCodoIPub = n.advertise<teleop_msgs::joint_angle>("LElbowRotation", 1);
  rotacionManoIPub = n.advertise<teleop_msgs::joint_angle>("LHandRotation", 1);
  aperturaManoIPub = n.advertise<std_msgs::Bool>("LHandOpening", 1);

  walkInfoPub = n.advertise<teleop_msgs::walk_info>("walkInfo", 1);

  ros::spin();
}
