#!/usr/bin/python

import math
import roslib; roslib.load_manifest('teleop_robot')
import rospy

from teleop_msgs.msg import joint_angle
from teleop_msgs.msg import walk_info
from std_msgs.msg import Bool
from std_msgs.msg import String

try:
	import naoqi
	import motion
	from naoqi import ALProxy
except ImportError:
	rospy.logerr("Error importing NaoQI. Please make sure that Aldebaran's NaoQI API is in your PYTHONPATH.")
	exit(1)



class ControlNao():
	def __init__(self, ip, port):
		rospy.init_node('teleop_nao')
		self.maxHeadSpeed = rospy.get_param('~max_head_speed', 0.3)

		self.motionProxy = ALProxy("ALMotion", ip , port)

		self.motionProxy.setStiffnesses("Body", 1.0)
		self.poseInicial()
		
		self.tts = ALProxy("ALTextToSpeech", ip, port)
		self.TO_RAD = math.pi/180
		

		self.listener()

	def poseInicial(self):
		import math
		TO_RAD = math.pi/180

		# Define The Initial Position
		HeadYawAngle       = + 0.0
		HeadPitchAngle     = + 0.0

		ShoulderPitchAngle = +80.0
		ShoulderRollAngle  = +20.0
		ElbowYawAngle      = -80.0
		ElbowRollAngle     = -60.0
		WristYawAngle      = + 0.0
		HandAngle          = + 0.0

		HipYawPitchAngle   = + 0.0
		HipRollAngle       = + 0.0
		HipPitchAngle      = -25.0
		KneePitchAngle     = +40.0
		AnklePitchAngle    = -20.0
		AnkleRollAngle     = + 0.0

		Head     = [HeadYawAngle, HeadPitchAngle]

		LeftArm  = [ShoulderPitchAngle, +ShoulderRollAngle, +ElbowYawAngle, +ElbowRollAngle, WristYawAngle, HandAngle]
		RightArm = [ShoulderPitchAngle, -ShoulderRollAngle, -ElbowYawAngle, -ElbowRollAngle, WristYawAngle, HandAngle]

		LeftLeg  = [HipYawPitchAngle, +HipRollAngle, HipPitchAngle, KneePitchAngle, AnklePitchAngle, +AnkleRollAngle]
		RightLeg = [HipYawPitchAngle, -HipRollAngle, HipPitchAngle, KneePitchAngle, AnklePitchAngle, -AnkleRollAngle]

		# Gather the joints together
		pTargetAngles = Head + LeftArm + LeftLeg + RightLeg + RightArm

		# Convert to radians
		pTargetAngles = [ x * motion.TO_RAD for x in pTargetAngles]

		#------------------------------ send the commands -----------------------------
		# We use the "Body" name to signify the collection of all joints
		pNames = "Body"
		
		# Ask motion to do this with a blocking call
		self.motionProxy.angleInterpolationWithSpeed(pNames, pTargetAngles, self.maxHeadSpeed)


	def nao_dice(self, data):
		try:
			self.tts.post.say(data.data)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)
        	

	def aperturaHombroDManejador (self,data):
		try:
			self.motionProxy.setAngles("RShoulderRoll", data.angle, self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def rotacionHombroDManejador (self,data):
		try:
			self.motionProxy.setAngles("RShoulderPitch", data.angle , self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)
		

	def aperturaCodoDManejador (self,data):
		try:
			self.motionProxy.setAngles("RElbowRoll", data.angle  , self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def rotacionCodoDManejador (self,data):
		try:
			self.motionProxy.setAngles("RElbowYaw", data.angle , self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def rotacionManoDManejador (self,data):
		try:
			self.motionProxy.setAngles("RWristYaw", data.angle, self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def aperturaManoDManejador (self,data):
		if data.data:
			try:
				self.motionProxy.openHand("RHand")
			except RuntimeError,e:
				rospy.logerr("Exception caught:\n%s", e)
		else:
			try:
				self.motionProxy.closeHand("RHand")
			except RuntimeError,e:
				rospy.logerr("Exception caught:\n%s", e)




	def rotacionCabezaManejador (self,data):
		try:
			self.motionProxy.setAngles("HeadYaw", data.angle, self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def inclinacionCabezaManejador (self,data):
		try:
			self.motionProxy.setAngles("HeadPitch", data.angle, self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)
		




	def aperturaHombroIManejador (self,data):
		try:
			self.motionProxy.setAngles("LShoulderRoll", data.angle, self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def rotacionHombroIManejador (self,data):
		try:
			self.motionProxy.setAngles("LShoulderPitch", data.angle , self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def aperturaCodoIManejador (self,data):
		try:
			self.motionProxy.setAngles("LElbowRoll", data.angle , self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def rotacionCodoIManejador (self,data):
		try:
			self.motionProxy.setAngles("LElbowYaw", data.angle, self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def rotacionManoIManejador (self,data):
		try:
			self.motionProxy.setAngles("LWristYaw", data.angle, self.maxHeadSpeed)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def aperturaManoIManejador (self,data):
		if data.data:
			try:
				self.motionProxy.openHand("LHand")
			except RuntimeError,e:
				rospy.logerr("Exception caught:\n%s", e)
		else:
			try:
				self.motionProxy.closeHand("LHand")
			except comparaRuntimeError,e:
				rospy.logerr("Exception caught:\n%s", e)

		


	def andadorManejador (self,data):
		try:
			self.motionProxy.setWalkTargetVelocity(data.x, data.y, data.angle, 1.0)
		except RuntimeError,e:
			rospy.logerr("Exception caught:\n%s", e)

	def listener(self):
	
		rospy.Subscriber("NaoSays", String, self.nao_dice, queue_size=1)

		rospy.Subscriber("RShoulderOpening", joint_angle, self.aperturaHombroDManejador, queue_size=1)
		rospy.Subscriber("RShoulderRotation", joint_angle, self.rotacionHombroDManejador, queue_size=1)
		rospy.Subscriber("RElbowOpening", joint_angle, self.aperturaCodoDManejador, queue_size=1)
		rospy.Subscriber("RElbowRotation", joint_angle, self.rotacionCodoDManejador, queue_size=1)
		rospy.Subscriber("RHandRotation", joint_angle, self.rotacionManoDManejador, queue_size=1)
		rospy.Subscriber("RHandOpening", Bool, self.aperturaManoDManejador, queue_size=1)

		rospy.Subscriber("HeadRotation", joint_angle, self.rotacionCabezaManejador, queue_size=1)
		rospy.Subscriber("HeadPitch", joint_angle, self.inclinacionCabezaManejador, queue_size=1)

		rospy.Subscriber("LShoulderOpening", joint_angle, self.aperturaHombroIManejador, queue_size=1)
		rospy.Subscriber("LShoulderRotation", joint_angle, self.rotacionHombroIManejador, queue_size=1)
		rospy.Subscriber("LElbowOpening", joint_angle, self.aperturaCodoIManejador, queue_size=1)
		rospy.Subscriber("LElbowRotation", joint_angle, self.rotacionCodoIManejador, queue_size=1)
		rospy.Subscriber("LHandRotation", joint_angle, self.rotacionManoIManejador, queue_size=1)
		rospy.Subscriber("LHandOpening", Bool, self.aperturaManoIManejador, queue_size=1)

		rospy.Subscriber("walkInfo", walk_info, self.andadorManejador, queue_size=1)
		rospy.spin()	

if __name__ == '__main__':

	# get connection from command line:
	from optparse import OptionParser

	rospy.sleep(5)

	parser = OptionParser()
	parser.add_option("--pip", dest="pip", default="127.0.0.1",
	                  help="IP/hostname of parent broker. Default is 127.0.0.1.", metavar="IP")
	parser.add_option("--pport", dest="pport", default=9559,
	                  help="port of parent broker. Default is 9559.", metavar="PORT")
		
	(options, args) = parser.parse_args()

	control = ControlNao(options.pip, int(options.pport))
	
	rospy.spin()
   	
	exit(0)
